from matplotlib.colors import ListedColormap

roseus: ListedColormap
roseus_r: ListedColormap
r: ListedColormap
r_r: ListedColormap
b: ListedColormap
b_r: ListedColormap
cyanus: ListedColormap
cyanus_r: ListedColormap
rbg: ListedColormap
rbg_r: ListedColormap
arcus: ListedColormap
arcus_r: ListedColormap
gr: ListedColormap
gr_r: ListedColormap
rg: ListedColormap
rg_r: ListedColormap
lavendula: ListedColormap
lavendula_r: ListedColormap
